"""Templates for the LLM Change Agent."""

from langchain_core.prompts.prompt import PromptTemplate


def get_issue_analyzer_template():
    """Issue analyzer template."""
    template = """
        {input}
        You are an semantic engineer. Based on the text you are given,
        you will analyze it.
        
        All that is expected of you is to form relevant KGCL commands in a list format.
        You have the following tools at your disposal to help you with this task:
        {tools}
        You also have the KGCL grammar in lark format: {grammar} along with an explanation of the grammar: {explanation}
        I want NO verbosity in your response. The final answer should consistently be a list of commands.

        Use the following format:

            Question: the input question you must answer
            Thought: you should always think about what to do
            Action: the action to take, should be one of [{tool_names}]
            Action Input: the input to the action
            Observation: the result of the action
            ... (this Thought/Action/Action Input/Observation can repeat N times)
            Thought: I now know the final answer
            Final Answer: the final answer to the original input question

            Begin!

            Question: {input}
            Thought:{agent_scratchpad}
    """
    return PromptTemplate(
        input_variables=[
            "input",
            "agent_scratchpad",
            "tools",
            "tool_names",
            "intermediate_steps",
            # "schema",
            "grammar",
            "explanation",
        ],
        template=template,
    )


def grammar_explanation():
    """Grammar explanation template."""
    return """
    The grammar defines commands for various operations such as renaming, creating, deleting, and modifying entities. It includes the following components:

    - `expression`: The entry point of the grammar, which can be any of the defined commands like `rename`, `create`, `delete`, etc.

    - `rename`: This command follows the pattern `"rename" _WS [id _WS "from" _WS] old_label ["@" old_language] _WS ("to"|"as"|"->") _WS new_label ["@" new_language]`.

    - `create`: This command follows the pattern `"create node" _WS id _WS label ["@" language]`.

    - `create_class`: This command follows the pattern `"create" _WS id`.

    - `create_synonym`: This command follows the pattern `"create" _WS [synonym_qualifier _WS] "synonym" _WS synonym ["@" language] _WS "for" _WS entity`.

    - `delete`: This command follows the pattern `"delete" _WS entity`.

    - `obsolete`: This command follows the pattern `"obsolete" _WS entity` or `"obsolete" _WS entity _WS "with replacement" _WS replacement`.

    - `unobsolete`: This command follows the pattern `"unobsolete" _WS entity`.

    - `deepen`: This command follows the pattern `"deepen" _WS entity _WS "from" _WS old_entity _WS ("to"|"->") _WS new_entity`.

    - `shallow`: This command follows the pattern `"shallow" _WS entity _WS "from" _WS old_entity _WS ("to"|"->") _WS new_entity`.

    - `move`: This command follows the pattern `"move" _WS entity_subject _WS entity_predicate _WS entity_object _WS "from" _WS old_entity _WS ("to"|"as"|"->") _WS new_entity`.

    - `create_edge`: This command follows the pattern `"create edge" _WS entity_subject _WS entity_predicate _WS entity_object_id`.

    - `delete_edge`: This command follows the pattern `"delete edge" _WS entity_subject _WS entity_predicate _WS entity_object_id`.

    - `change_relationship`: This command follows the pattern `"change relationship between" _WS entity_subject _WS "and" _WS entity_object _WS "from" _WS old_entity _WS "to" _WS new_entity`.

    - `change_annotation`: This command follows the pattern `"change annotation of" _WS entity_subject _WS "with" _WS entity_predicate _WS "from" _WS old_entity_object _WS "to" _WS new_entity_object`.

    - `change_definition`: This command follows the pattern `"change definition of" _WS entity _WS "to" _WS new_definition` or `"change definition of" _WS entity _WS "from" _WS old_definition _WS "to" _WS new_definition`.

    - `add_definition`: This command follows the pattern `"add definition" _WS new_definition _WS "to" _WS entity`.

    - `remove_definition`: This command follows the pattern `"remove definition for" _WS entity`.

    - `remove_from_subset`: This command follows the pattern `"remove" _WS id _WS "from subset" _WS subset`.

    - `add_to_subset`: This command follows the pattern `"add" _WS id _WS "to subset" _WS subset`.

    The `%import` statements import common definitions for `ID`, `LABEL`, `CURIE`, `SINGLE_QUOTE_LITERAL`, `TRIPLE_SINGLE_QUOTE_LITERAL`, `DOUBLE_QUOTE_LITERAL`, `TRIPLE_DOUBLE_QUOTE_LITERAL`, `LANGUAGE_TAG`, and whitespace (`_WS`). The `%ignore` statement tells the parser to ignore whitespace.

    """



"""Constants for the LLM Change Agent."""

from os import getenv

from importlib_metadata import files

OPENAI_KEY = str(getenv("OPENAI_API_KEY"))
ANTHROPIC_KEY = str(getenv("ANTHROPIC_API_KEY"))
CBORG_KEY = str(getenv("CBORG_API_KEY"))

OPEN_AI_MODEL = "gpt-4o-2024-08-06"
ANTHROPIC_MODEL = "claude-3-5-sonnet-20240620"
OLLAMA_MODEL = "llama3.1"  #!  not all models support tools (tool calling)
CBORG_MODEL = "lbl/llama-3"

KGCL_SCHEMA = [file for file in files("kgcl-schema") if file.stem == "kgcl" and file.suffix == ".yaml"][0]
KGCL_GRAMMAR = [file for file in files("kgcl-schema") if file.stem == "kgcl" and file.suffix == ".lark"][0]



"""Configuration classes for the LLM models."""

from pydantic import BaseModel


class LLMConfig(BaseModel):
    """Configuration for the LLM model."""

    model: str
    provider: str
    temperature: float = 0.0


class OpenAIConfig(LLMConfig):
    """Configuration for OpenAI LLM model."""

    pass


class OllamaConfig(LLMConfig):
    """Configuration for Ollama LLM model."""

    format: str = None
    pass


class AnthropicConfig(LLMConfig):
    """Configuration for Anthropic LLM model."""

    pass


class CBORGConfig(LLMConfig):
    """Configuration for CBORG LLM model."""

    base_url: str = "https://api.cborg.lbl.gov"  # Local clients can also use https://api-local.cborg.lbl.gov



"""Utility functions for the LLM Change Agent."""

import yaml
from langchain.agents import AgentExecutor
from langchain.agents.react.agent import create_react_agent
from langchain.tools.retriever import create_retriever_tool
from langchain_chroma import Chroma
from langchain_core.documents import Document
from langchain_openai import OpenAIEmbeddings
from langchain_text_splitters import RecursiveCharacterTextSplitter
from openai import OpenAI

from llm_change_agent.config.llm_config import AnthropicConfig, CBORGConfig, LLMConfig, OllamaConfig, OpenAIConfig
from llm_change_agent.constants import ANTHROPIC_KEY, CBORG_KEY, KGCL_GRAMMAR, KGCL_SCHEMA, OPENAI_KEY
from llm_change_agent.templates.templates import get_issue_analyzer_template, grammar_explanation


def get_openai_models():
    """Get the list of OpenAI models."""
    sorted_model_ids = []
    if OPENAI_KEY != "None":
        openai = OpenAI()
        models_list = sorted(
            [model for model in openai.models.list() if model.id.startswith("gpt-4") and model.created >= 1706037777],
            key=lambda x: x.created,
            reverse=True,
        )
        sorted_model_ids = [model.id for model in models_list]
    return sorted_model_ids


def get_anthropic_models():
    """Get the list of Anthropic models."""
    return [
        "claude-3-5-sonnet-20240620",
        "claude-3-opus-20240229",
        "claude-3-sonnet-20240229",
        "claude-3-haiku-20240307",
    ]


def get_ollama_models():
    """Get the list of Ollama models."""
    return ["llama3.1"]


def get_lbl_cborg_models():
    """Get the list of LBNL-hosted models via CBORG."""
    return [
        "lbl/llama-3",  # LBNL-hosted model (free to use)
        "openai/chatgpt:latest",  # OpenAI-hosted model
        "anthropic/claude:latest",  # Anthropic-hosted model
        "google/gemini:latest",  # Google-hosted model
    ]


def get_provider_model_map():
    """Get the provider to model mapping."""
    return {
        "openai": get_openai_models(),
        "ollama": get_ollama_models(),
        "anthropic": get_anthropic_models(),
        "cborg": get_lbl_cborg_models(),
    }


def get_provider_for_model(model):
    """Get the provider for the model."""
    provider_model_map = get_provider_model_map()
    for provider, models in provider_model_map.items():
        if model in models:
            return provider
    return None


def get_default_model_for_provider(provider):
    """Get the default model for the provider."""
    provider_model_map = get_provider_model_map()
    if provider in provider_model_map:
        return provider_model_map[provider][0]
    return None


def get_api_key(provider):
    """Get the API key for the provider."""
    if provider == "openai":
        return OPENAI_KEY
    elif provider == "anthropic":
        return ANTHROPIC_KEY
    elif provider == "cborg":
        return CBORG_KEY
    return None


def llm_factory(config: LLMConfig):
    """Create an LLM instance based on the configuration."""
    if isinstance(config, OpenAIConfig):
        from langchain_openai import ChatOpenAI

        return ChatOpenAI(model=config.model, temperature=config.temperature, api_key=get_api_key(config.provider))
    elif isinstance(config, OllamaConfig):
        from langchain_ollama import ChatOllama

        return ChatOllama(model=config.model, temperature=config.temperature, api_key=get_api_key(config.provider))
    elif isinstance(config, AnthropicConfig):
        from langchain_anthropic import ChatAnthropic

        return ChatAnthropic(
            model=config.model,
            temperature=config.temperature,
            api_key=get_api_key(config.provider),
            max_tokens_to_sample=4096,
        )
    elif isinstance(config, CBORGConfig):
        from langchain_openai import ChatOpenAI

        return ChatOpenAI(
            model=config.model,
            temperature=config.temperature,
            openai_api_key=get_api_key(config.provider),
            openai_api_base=config.base_url,
        )

    else:
        raise ValueError("Unsupported LLM configuration")


def get_kgcl_schema():
    """Get the KGCL schema information."""
    with open(KGCL_SCHEMA.locate(), "r") as schema_yaml:
        schema = yaml.safe_load(schema_yaml)
    return schema


def get_kgcl_grammar():
    """Get the KGCL grammar information."""
    with open(KGCL_GRAMMAR.locate(), "r") as grammar_file:
        lark_file = grammar_file.read()
    grammar_notes = grammar_explanation()
    return {"lark": lark_file, "explanation": grammar_notes}


def split_documents(document: str):
    """Split the document into a list of documents."""
    doc_object = (Document(page_content=document),)
    splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=200)
    splits = splitter.split_documents(doc_object)
    return splits


def execute_agent(llm, prompt):
    """Create a retriever agent."""
    grammar = get_kgcl_grammar()
    # schema = get_kgcl_schema()
    # docs_list = (
    #     split_documents(str(schema)) + split_documents(grammar["lark"]) + split_documents(grammar["explanation"])
    # )

    docs_list = split_documents(grammar["lark"]) + split_documents(grammar["explanation"])
    vectorstore = Chroma.from_documents(documents=docs_list, embedding=OpenAIEmbeddings())
    retriever = vectorstore.as_retriever(search_kwargs={"k": 1})
    tool = create_retriever_tool(retriever, "change_agent_retriever", "Change Agent Retriever")
    tools = [tool]
    template = get_issue_analyzer_template()
    react_agent = create_react_agent(llm=llm, tools=tools, prompt=template)
    agent_executor = AgentExecutor(agent=react_agent, tools=tools, handle_parsing_errors=True, verbose=True)

    return agent_executor.invoke(
        {
            "input": prompt,
            # "schema": schema,
            "grammar": grammar["lark"],
            "explanation": grammar["explanation"],
        }
    )


def augment_prompt(prompt: str):
    """Augment the prompt with additional information."""
    return f"""
        Give me all relevant KGCL commands based on this request: \n\n
        + {prompt} +
        \n\nReturn as a JSON format list.\n\n"""



"""Main python file."""

from pprint import pprint

from llm_change_agent.constants import OPEN_AI_MODEL
from llm_change_agent.utils.llm_utils import (
    augment_prompt,
    execute_agent,
    get_anthropic_models,
    get_default_model_for_provider,
    get_lbl_cborg_models,
    get_ollama_models,
    get_openai_models,
    get_provider_for_model,
    get_provider_model_map,
    llm_factory,
)


class LLMChangeAgent:
    """Define LLMChangeAgent class."""

    def __init__(self, model: str, prompt: str, provider: str):
        """Initialize LLMChangeAgent class."""
        self.model = model
        self.prompt = prompt
        self.provider = provider
        self.llm = None

    def _get_llm_config(self):
        """Get the LLM configuration based on the selected LLM."""

        def _validate_and_get_model(llm_model, get_models_func, default_model=OPEN_AI_MODEL):
            """Validate the model and get the model."""
            if llm_model is None:
                if self.provider is None:
                    from .config.llm_config import OpenAIConfig

                    return OpenAIConfig(model=default_model, provider=get_provider_for_model(default_model))
                else:
                    llm_model = get_default_model_for_provider(self.provider)

            list_of_models = get_models_func()
            if llm_model not in list_of_models:
                raise ValueError(f"Model {llm_model} not supported. Please choose from {list_of_models}")
            return llm_model

        provider_config_map = {
            "openai": ("OpenAIConfig", get_openai_models),
            "ollama": ("OllamaConfig", get_ollama_models),
            "anthropic": ("AnthropicConfig", get_anthropic_models),
            "cborg": ("CBORGConfig", get_lbl_cborg_models),
        }

        if self.provider in provider_config_map:
            config_class_name, get_models_func = provider_config_map[self.provider]
            llm_model = _validate_and_get_model(self.model, get_models_func)
            config_module = __import__("llm_change_agent.config.llm_config", fromlist=[config_class_name])
            ConfigClass = getattr(config_module, config_class_name)
            return ConfigClass(model=llm_model, provider=get_provider_for_model(llm_model))

        else:
            all_models = get_provider_model_map()
            if self.model is None:
                from .config.llm_config import OpenAIConfig

                return OpenAIConfig(model=OPEN_AI_MODEL, provider=get_provider_for_model(OPEN_AI_MODEL))

            for provider, models in all_models.items():
                if self.model in models:
                    # Temporarily set the provider and model to use the existing logic
                    original_provider = self.provider
                    original_model = self.model
                    self.provider = provider
                    self.model = self.model
                    try:
                        return self._get_llm_config()
                    finally:
                        # Restore the original provider and model
                        self.provider = original_provider
                        self.model = original_model

            raise ValueError(f"Model {self.model} not supported.")

    def run(self):
        """Run the LLM Change Agent."""
        llm_config = self._get_llm_config()
        self.llm = llm_factory(llm_config)
        response = execute_agent(llm=self.llm, prompt=augment_prompt(self.prompt))
        pprint(response["output"])
        return response["output"]



"""Command line interface for llm-change-agent."""

import logging

import click

from llm_change_agent import __version__
from llm_change_agent.llm_agent import LLMChangeAgent
from llm_change_agent.utils.llm_utils import (
    get_anthropic_models,
    get_lbl_cborg_models,
    get_ollama_models,
    get_openai_models,
)

ALL_AVAILABLE_PROVIDERS = ["openai", "ollama", "anthropic", "cborg"]
ALL_AVAILABLE_MODELS = get_openai_models() + get_ollama_models() + get_anthropic_models() + get_lbl_cborg_models()

__all__ = [
    "main",
]

logger = logging.getLogger(__name__)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.version_option(__version__)
def main(verbose: int, quiet: bool):
    """
    CLI for llm-change-agent.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)


@main.command()
def list_models():
    """List the available language models."""
    openai_models = "\n  ".join(get_openai_models())
    anthropic_models = "\n  ".join(get_anthropic_models())
    ollama_models = "\n  ".join(get_ollama_models())
    lbl_cborg_models = "\n  ".join(get_lbl_cborg_models())

    click.echo(f"OpenAI models:\n  {openai_models}")
    click.echo(f"Anthropic models:\n  {anthropic_models}")
    click.echo(f"Ollama models:\n  {ollama_models}")
    click.echo(f"LBL-CBORG models:\n  {lbl_cborg_models}")


@main.command()
@click.option("--model", type=click.Choice(ALL_AVAILABLE_MODELS), help="Model to use for generation.")
@click.option("--provider", type=click.Choice(ALL_AVAILABLE_PROVIDERS), help="Provider to use for generation.")
@click.option("--prompt", type=str, default="Hello, world!", help="Prompt to use for generation.")
def execute(model: str, provider: str, prompt: str):
    """Generate text using the specified model."""
    llm_agent = LLMChangeAgent(model=model, prompt=prompt, provider=provider)
    llm_agent.run()


if __name__ == "__main__":
    main()
